import sammy as sm


sam = sm.SAM(Description='Simple CRUD webservice. State is stored in a SimpleTable (DynamoDB) resource.')

sam.add_resource(sm.Function(
    name='GetFunction',
    CodeUri=sm.S3URI(Bucket='<bucket>',Key='sammytest.zip'),
    Handler='index.get',
    Runtime='python3.6',
    Policies='AmazonDynamoDBReadOnlyAccess',
    Environment=sm.Environment(Variables={'TABLE_NAME':sm.Ref(Ref='Table')}),
    Events=[sm.APIEvent(name='GetResource',Path='/resource/{resourceId}',Method='get')]
))

sam.add_resource(sm.Function(
    name='PutFunction',
    Handler='index.put',
    Runtime='python3.6',
    CodeUri=sm.S3URI(Bucket='<bucket>',Key='sammytest.zip'),
    Policies='AmazonDynamoDBFullAccess',
    Environment=sm.Environment(Variables={'TABLE_NAME':sm.Ref(Ref='Table')}),
    Events=[sm.APIEvent(name='PutResource',Path='/resource/{resourceId}',Method='put')]
))

sam.add_resource(sm.Function(
    name='DeleteFunction',
    Handler='index.delete',
    Runtime='python3.6',
    CodeUri=sm.S3URI(Bucket='<bucket>',Key='sammytest.zip'),
    Policies='AmazonDynamoDBFullAccess',
    Environment=sm.Environment(Variables={'TABLE_NAME':sm.Ref(Ref='Table')}),
    Events=[sm.APIEvent(name='DeleteResource',Path='/resource/{resourceId}',Method='delete')]
))

sam.add_resource(sm.SimpleTable(name='Table'))


import sammy as sm


sam = sm.SAM(Description='Alexa Skill https://developer.amazon.com/alexa-skills-kit')

sam.add_resource(sm.Function(
    name='AlexaSkillFunction',
    CodeUri=sm.S3URI(Bucket='<bucket>',Key='sammytest.zip'),
    Handler='sample.handler',
    Runtime='python3.6',
    Events=[sm.AlexaSkillEvent(name='AlexaSkillEvent')]
))



import json

from valley.exceptions import ValidationException
from valley.mixins import ListMixin
from valley.properties import BaseProperty, MultiProperty
from valley.utils.json_utils import ValleyEncoder
from valley.validators import ForeignListValidator, StringValidator, ForeignValidator


class ForeignSubclassListValidator(ForeignListValidator):

    def validate(self, value, key):
        if value:
            for obj in value:
                if not issubclass(obj.__class__,self.foreign_class):
                    raise ValidationException(
                        '{0}: This value ({1}) should be an instance of {2}.'.format(
                            key, obj, self.foreign_class.__name__)
                    )



class ForeignSubclassListMixin(ListMixin):

    def get_validators(self):
        super(ForeignSubclassListMixin, self).get_validators()
        self.validators.insert(len(self.validators),ForeignSubclassListValidator(self.foreign_class))

    def get_db_value(self, value):
        if self.return_type == 'single':
            if not self.return_prop:
                raise ValueError('ForeignProperty classes requires the '
                    'return_prop argument if return_type equals "single"')
            return value._data[self.return_prop]
        if self.return_type == 'list':
            return [obj._data for obj in value]
        if self.return_type == 'json':
            return json.dumps(value, cls=ValleyEncoder)
        else:
            return value


class ForeignInstanceListProperty(ForeignSubclassListMixin,BaseProperty):

    def __init__(self,foreign_class,return_type=None,return_prop=None,**kwargs):
        self.foreign_class = foreign_class
        super(ForeignInstanceListProperty, self).__init__(**kwargs)
        self.return_type = return_type
        self.return_prop = return_prop


class CharForeignProperty(MultiProperty):

    def __init__(self,foreign_class,**kwargs):

        super(CharForeignProperty, self).__init__(
            validators=[ForeignValidator(foreign_class),
                        StringValidator()],**kwargs)





import sammy as sm


sam = sm.SAM(Description='A hello world application.',render_type='yaml')

sam.add_parameter(sm.Parameter(name='Bucket',Type='String'))

sam.add_parameter(sm.Parameter(name='CodeZipKey',Type='String'))

sam.add_resource(
    sm.Function(name='HelloWorldFunction',
        Handler='sample.handler', Runtime='python3.6', CodeUri=sm.S3URI(
            Bucket=sm.Ref(Ref='Bucket'),Key=sm.Ref(Ref='CodeZipKey'))))



import boto3
import json
import yaml

from awscli.customizations.cloudformation.deployer import Deployer
from valley.properties import *
from valley.contrib import Schema
from valley.utils.json_utils import ValleyEncoderNoType

from sammy.custom_properties import ForeignInstanceListProperty, CharForeignProperty


API_METHODS = {
    'post':'post',
    'get':'get',
    'head':'head',
    'delete':'delete',
    'put':'put',
    'options':'options',
    'connect':'connect',
    'any':'any'
}
RENDER_FORMATS = {'json':'json','yaml':'yaml'}

def remove_nulls(obj_dict):
    null_keys = []
    for k, v in obj_dict.items():
        if not v:
            null_keys.insert(0, k)
    for k in null_keys:
        obj_dict.pop(k)
    return obj_dict


class SAMSchema(Schema):

    def __init__(self,**kwargs):
        super(SAMSchema, self).__init__(**kwargs)
        self.validate()


class Ref(SAMSchema):
    Ref = CharProperty(required=True)


class S3URI(SAMSchema):
    Bucket = CharForeignProperty(Ref,required=True)
    Key = CharForeignProperty(Ref,required=True)

    def to_dict(self):
        obj = remove_nulls(self._data.copy())
        return obj


class S3KeyFilter(SAMSchema):
    S3Key = CharForeignProperty(Ref)


class Environment(SAMSchema):
    Variables = DictProperty(required=True)


class Parameter(SAMSchema):
    name = CharForeignProperty(Ref,required=True)
    Type = CharForeignProperty(Ref,required=True)

    def to_dict(self):
        obj = remove_nulls(self._data.copy())
        name = obj.pop('name')
        return {
            'name':name,
            'r':obj
        }

class Resource(SAMSchema):
    _resource_type = None

    name = CharForeignProperty(Ref,required=True)

    def to_dict(self):
        obj = remove_nulls(self._data.copy())
        name = obj.pop('name')
        r_attrs = {
            'Type':self._resource_type
        }
        if len(obj.keys()) > 0:
            r_attrs['Properties'] = {k: v for k, v in obj.items() if v}
        return {
            'name':name,
            'r':r_attrs
        }


    def add_attr(self,k,v):
        self.r_attrs['Properties'][k] = v


class EventSchema(SAMSchema):
    _event_type = None

    name = CharForeignProperty(Ref,required=True)

    def to_dict(self):
        obj = remove_nulls(self._data.copy())
        event = {'name':obj.pop('name'),
                 'r':{
                    'Type':self._event_type
                 }
                }

        if len(obj.keys()) > 0:
            event['r']['Properties'] = obj
        return event


class APIEvent(EventSchema):
    _event_type = 'Api'

    Path = CharForeignProperty(Ref,required=True)
    Method = CharForeignProperty(Ref,required=True,choices=API_METHODS)
    RestApiId = CharForeignProperty(Ref)


class S3Event(EventSchema):
    _event_type = 'S3'

    Bucket = CharForeignProperty(Ref,required=True)
    Events = ListProperty(required=True)
    Filter = ForeignProperty(S3KeyFilter)


class SNSEvent(EventSchema):
    _event_type = 'SNS'

    Topic = CharForeignProperty(Ref,required=True)


class KinesisEvent(EventSchema):
    _event_type = 'Kinesis'

    Stream = CharForeignProperty(Ref,required=True)
    StartingPosition = CharForeignProperty(Ref,required=True)
    BatchSize = IntegerProperty()


class DynamoDBEvent(EventSchema):
    _event_type = 'DynamoDB'

    Stream = CharForeignProperty(Ref,required=True)
    StartingPosition = CharForeignProperty(Ref,required=True)
    BatchSize = IntegerProperty()


class ScheduleEvent(EventSchema):
    Schedule = CharForeignProperty(Ref,required=True)
    Input = CharForeignProperty(Ref)


class CloudWatchEvent(EventSchema):
    Pattern = DictProperty(required=True)
    Input = CharForeignProperty(Ref)
    InputPath = CharForeignProperty(Ref)


class IoTRuleEvent(EventSchema):
    Sql = CharForeignProperty(Ref,required=True)
    AwsIotSqlVersion = CharForeignProperty(Ref)


class AlexaSkillEvent(EventSchema):
    _event_type = 'AlexaSkill'


class DeadLetterQueueSchema(SAMSchema):
    _dlq_type = None

    name = CharForeignProperty(Ref,required=True)
    TargetArn = CharForeignProperty(Ref,required=True)

    def to_dict(self):
        obj = remove_nulls(self._data.copy())
        event = {'name':obj.pop('name'),
                 'r':{
                    'Type':self._dlq_type,
                    'Properties':obj
                }}

        return event


class SNSLetterQueue(DeadLetterQueueSchema):
    _dlq_type = 'SNS'


class SQSLetterQueue(DeadLetterQueueSchema):
    _dlq_type = 'SQS'


class Function(Resource):
    _resource_type = 'AWS::Serverless::Function'

    Handler = CharForeignProperty(Ref,required=True)
    Runtime = CharForeignProperty(Ref,required=True,max_length=15)
    CodeUri = ForeignProperty(S3URI)
    FunctionName = CharForeignProperty(Ref)
    Description = CharForeignProperty(Ref)
    MemorySize = IntegerProperty()
    Timeout = IntegerProperty()
    Role = CharForeignProperty(Ref)
    Policies = CharForeignProperty(Ref)
    Environment = ForeignProperty(Environment)
    VpcConfig = DictProperty()
    Events = ForeignInstanceListProperty(EventSchema)
    Tags = DictProperty()
    Tracing = CharForeignProperty(Ref)
    KmsKeyArn = CharForeignProperty(Ref)
    DeadLetterQueue = ForeignInstanceListProperty(DeadLetterQueueSchema)

    def to_dict(self):

        obj = super(Function, self).to_dict()
        try:
            events = [i.to_dict() for i in obj['r']['Properties'].pop('Events')]
            obj['r']['Properties']['Events'] = {i.get('name'):i.get('r') for i in events}
        except KeyError:
            pass

        try:
            dlq = [i.to_dict() for i in obj['r']['Properties'].pop('DeadLetterQueue')]
            obj['r']['Properties']['DeadLetterQueue'] = {i.get('name'):i.get('r') for i in dlq}
        except KeyError:
            pass

        return obj


class API(Resource):
    _resource_type = "AWS::Serverless::Api"

    StageName = CharForeignProperty(Ref,required=True)
    DefinitionUri = CharForeignProperty(Ref)
    DefinitionBody = DictProperty()
    CacheClusterEnabled = BooleanProperty()
    CacheClusterSize = CharForeignProperty(Ref)
    Variables = DictProperty()


class SimpleTable(Resource):
    _resource_type = "AWS::Serverless::SimpleTable"

    PrimaryKey = DictProperty()
    ProvisionedThroughput = DictProperty()


class SAM(SAMSchema):
    aws_template_format_version = '2010-09-09'
    transform = 'AWS::Serverless-2016-10-31'
    Description = CharProperty()
    resources = ForeignInstanceListProperty(Resource)
    parameters = ForeignInstanceListProperty(Parameter)
    render_type = CharProperty(choices=RENDER_FORMATS,default_value='yaml')

    def __init__(self,**kwargs):
        super(SAM, self).__init__(**kwargs)

        self.cf = boto3.client('cloudformation')
        self.s3 = boto3.resource('s3')

    def add_parameter(self,parameter):
        self._base_properties.get('parameters').validate([parameter],'parameters')
        parameters = self._data.get('parameters') or []
        parameters.append(parameter)
        parameters = set(parameters)
        self._data['parameters'] = list(parameters)

    def add_resource(self,resource):
        self._base_properties.get('resources').validate([resource],'resources')
        resources = self._data.get('resources') or []
        resources.append(resource)
        resources = set(resources)
        self._data['resources'] = list(resources)

    def to_dict(self):
        obj = remove_nulls(self._data.copy())
        rl = [i.to_dict() for i in obj.get('resources')]

        resources = {i.get('name'):i.get('r') for i in rl}


        template = {
            'AWSTemplateFormatVersion': self.aws_template_format_version,
            'Transform': self.transform,
            'Resources': resources
        }
        if obj.get('Description'):
            template['Description'] = obj.get('Description')
        if obj.get('parameters'):
            pl = [i.to_dict() for i in obj.get('parameters')]
            parameters = {i.get('name'): i.get('r') for i in pl}
            if len(parameters.keys()) > 0:
                template['Parameters'] = parameters
        return template

    def get_template_dict(self):
        return self.to_dict()

    def publish_template(self,bucket,name):
        filename = '{}.{}'.format(name,self.render_type)

        self.s3.Object(bucket, filename).put(
            Body=self.get_template())

    def get_template(self):
        if self.render_type == 'json':
            return self.to_json()
        else:
            return self.to_yaml()

    def publish(self, stack_name,**kwargs):
        param_list = [{'ParameterKey':k,'ParameterValue':v} for k,v in kwargs.items()]
        d = Deployer(boto3.client('cloudformation'))
        result = d.create_and_wait_for_changeset(
            stack_name=stack_name,
            cfn_template=self.get_template(),
            parameter_values=param_list,
            capabilities=['CAPABILITY_IAM'])
        d.execute_changeset(result.changeset_id, stack_name)
        d.wait_for_execute(stack_name, result.changeset_type)

    def to_yaml(self):
        jd = json.dumps(self.get_template_dict(),cls=ValleyEncoderNoType)
        #TODO: Write this without converting to JSON first
        jl = json.loads(jd)
        return yaml.safe_dump(jl,
                              default_flow_style=False)

    def to_json(self):
        return json.dumps(self.get_template_dict(),cls=ValleyEncoderNoType)


from setuptools import setup, find_packages
from pip.req import parse_requirements

install_reqs = parse_requirements('requirements.txt', session=False)

version = '0.2.0'

setup(
    name='sammy',
    version=version,
    description="Python library for generating AWS SAM "
                "(Serverless Application Model) templates with validation.",
    classifiers=[
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Environment :: Web Environment",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3 :: Only"
    ],
    keywords='serverless, cloudformation, sam',
    author='Brian Jinwright',
    author_email='opensource@ipoots.com',
    maintainer='Brian Jinwright',
    packages=find_packages(),
    url='https://github.com/capless/sammy',
    license='GNU General Public License v3.0',
    install_requires=[str(ir.req) for ir in install_reqs],
    include_package_data=True,
    zip_safe=False,
)



import unittest
import pathlib as pl
import os
import yaml

from sammy.examples.alexa_skill import sam as al
from sammy.examples.api_backend import sam as ab
from sammy.examples.hello_world import sam as hw


class AlexaTestCase(unittest.TestCase):
    template_name = 'alexa_skill.yaml'
    template = al

    def setUp(self):
        template_path = '{}/{}'.format(
            pl.PurePath(
                os.path.abspath(__file__)).parent / 'examples/yaml', self.template_name)
        with open(template_path, 'r') as f:
            self.template_dict = yaml.load(f)

    def test_template(self):
        s = yaml.safe_load(self.template.get_template())
        self.assertEqual(self.template_dict,s)


class APIBackendTestCase(AlexaTestCase):
    template_name = 'api_backend.yaml'
    template = ab


class HelloWorldTestCase(AlexaTestCase):
    template_name = 'hello_world.yaml'
    template = hw